local nametag, addon = ...

local empty_text = "No alternative currencies are being tracked.  Click "..
    "on specific token types in the currency window to track them."
local currency_checkbox_tooltip_text = "Track this currency across characters in Gold Watch."

addon.defaults = {
	global = {
		useIcons = true,
		iconSize = 20,
		shortcuts = true,
		tracked = {},
	},
	factionrealm = {
	},
}


-- NOTA BENE:  If graphical elements are added, increase the vertical size
-- being set at the comment 'OPTIONS TABLE HEIGHT'
addon.options = {
	name = "Gold Watch",
	type = 'group',
	handler = addon,  -- functions listed as strings called as addon:function
	args = {
		win = {
			name = "show window",
			desc = "Open options window",
			type = 'execute',
			guiHidden = true,
			func = function(info)
				InterfaceOptionsFrame_OpenToCategory(info.handler.ioptionsWidget.frame)
			end
		},
		version = {
			--name = filled in below
			type = 'description',
			cmdHidden = true,
			order = 10,
			width = 'full',
		},
		reset = {
			name = "Reset money counter",
			desc = "Resets earned/lost counter to zero.  Shortcut: '/gw reset'",
			type = 'execute',
			order = 20,
			func = function(info)
				addon:gwReset()
				addon:Print"GoldWatch delta counter reset"
			end,
		},
		_divider1 = {
			name = "",
			type = 'header',
			cmdHidden = true,
			order = 30,
		},
		useIcons = {
			name = "Use icons",
			desc = "Whether to use icons or single letters in chat output",
			type = 'toggle',
			order = 40,
			get = "GetOption", set = "SetOption",
		},
		iconSize = {
			name = "Icon size",
			desc = "Size of coin icon in chat output",
			type = 'range',
			order = 50,
			min = 6,
			max = 30,
			step = 1,
			get = "GetOption", set = "SetOption",
		},
		shortcuts = {
			name = "Register shortcut /slashes",
			desc = "Whether to register shortcut slash commands (/gw, /tw) when logging in",
			type = 'toggle',
			order = 60,
			width = 'double',  -- else desc field is truncated
			get = "GetOption", set = "SetOption",
		},

		_divider2 = {
			name = "",
			type = 'header',
			cmdHidden = true,
			order = 10001,
		},
		_blurb = {
			--name = filled in below
			type = 'description',
			cmdHidden = true,
			order = 10002,
		},
		_divider3 = {
			name = "",
			type = 'header',
			cmdHidden = true,
			order = 10003,
		},

		-- now the "long cuts", for command-line use only
		status = {
			name = "Counter status",
			desc = "Display money earned or lost.  Shortcut: '/gw'",
			type = 'execute',
			order = 200,
			guiHidden = true,
			func = 'gwStatus',
		},
		all = {
			name = "Total",
			desc = "Display total money across characters on this faction.  Shortcut: '/gw all'",
			type = 'execute',
			order = 210,
			guiHidden = true,
			func = 'gwAll', arg = false
		},
		["all+"] = {
			name = "Total broken down",
			desc = "Same as 'all' but split up by character.  Shortcut: '/gw all+'",
			type = 'execute',
			order = 211,
			guiHidden = true,
			func = 'gwAll', arg = true
		},
	}
}

do
	local version = "<Version " .. (GetAddOnMetadata(nametag, "Version") or "?")
		.. " wow.curse.com, by Farmbuyer+Elennar@US-Kilrogg>"
	local blurb = "All options and money displays are available through the "..
		"</goldwatch> command.  Use </goldwatch> by itself for a complete listing.  "..
		"This window controls options only.\n\n"..
		"Quick access is through the </gw> and </tw> commands (see </goldwatch> "..
		"for the shortcuts)."
	local blurb_colors = { ["<"]="|cff00ff00", [">"]="|r" }

	addon.options.args.version.name = version:gsub("[<>]", blurb_colors)
	addon.options.args._blurb.name = blurb:gsub("[<>]", blurb_colors)
end


-----------------------------------------------------------------------------
-- other locals
local globalSV, shareSV, mySV

-- smartGetItemIcon returns the icon texture string for item IDs, but
-- reproduces the TokenUI behavior for special currency tokens.  (Else we get
-- weird silly tokens for honor/arena points.)  Also returns TexCoord
-- numbers, for feeding directly into acegui's SetImage routine.  The data
-- are cached to try and speed things up.
local cache_smartGetItemIcon, smartGetItemIcon
do
	local sGII = {}
	local function cache_smartGetItemIcon (id, extratype)
		if sGII[id] then return end
		local v
		if extratype == 1 then  -- arena
			v = { "Interface\\PVPFrame\\PVP-ArenaPoints-Icon", 0, 1, 0, 1 }
		elseif extratype == 2 then  -- honor
			local factionGroup = UnitFactionGroup("player")
			if factionGroup then
				v = { "Interface\\TargetingFrame\\UI-PVP-"..factionGroup, 0.03125, 0.59375, 0.03125, 0.59375 }
			else -- how the hell did this happen? fall back to arena icon
				v = { "Interface\\PVPFrame\\PVP-ArenaPoints-Icon", 0, 1, 0, 1 }
			end
		else
			v = { GetItemIcon(id), 0, 1, 0, 1 }
		end
		sGII[id] = v
	end
	local function smartGetItemIcon (id)
		if not sGII[id] then 
			cache_smartGetItemIcon(id)
		end
		return unpack(sGII[id])
	end
end

-- Indexed by "token ID" (the index of the list of token currencies on the
-- character currency tab), maps to the itemID for that token.
local token_map = {}

-- The idea here is to tell the user what's changed, but delay printing the
-- value until they're finished.
local options_set_haschanged = {}
local function options_set_informuser()
	for optionname in pairs(options_set_haschanged) do
		addon:Printf("|cff5555ff%s|rnow set to %s", optionname, globalSV[optionoption])
	end
	table.wipe(options_set_haschanged)
end

-- Iterate a table T according to its entries' CURRENCY key.  Based on
-- the standard example from PiL ed1 19.3.  Loop variables are T's
-- entries and those entries' CURRENCY values.
local function sorted_currency_pairs (t, currency)
	local a = {}
	for name,counts in pairs(t) do
		if counts[currency] then
			table.insert(a, {name, counts[currency]})
		end
	end
	table.sort(a, function(L,R) return L[2] > R[2] end)
	local i = 0				-- iterator state
	local iter = function()	-- iterator
		i = i + 1
		if a[i] == nil then return nil
		elseif a[i][2] == 0 then return nil  -- XXX maybe...
		else return a[i][1], a[i][2]
		end
	end
	return iter
end




-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, nametag,
		        "AceConsole-3.0", "AceTimer-3.0", "AceEvent-3.0", "AceHook-3.0")

function addon:SetOption (info, value)
	local name = info[#info]
	self.db.global[name] = value
	--local arg = info.arg
	--if arg then self[arg](self) end
	options_set_haschanged[name] = true
	if info.uiType == "cmd" then options_set_informuser() end   -- not 'dialog', so give immediate feedback
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.global[name]
end

function addon:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("GoldWatchDB", self.defaults, --[[Default=]]true)
	local me = UnitName('player')
	if type(self.db.factionrealm[me]) ~= 'table' then
		self.db.factionrealm[me] = {}
	end
	globalSV, shareSV, mySV = self.db.global, self.db.factionrealm, self.db.factionrealm[me]

	LibStub("AceConfig-3.0"):RegisterOptionsTable(nametag, self.options, "goldwatch")

	if globalSV.shortcuts then
		self:RegisterChatCommand("gw", "gwSlash")
		self:RegisterChatCommand("tw", "twSlash")
	end
	--self:RegisterChatCommand("Q", "hack")
	self.not_yet_scanned = true

	self.OnInitialize = nil
end
--function addon:hack() InterfaceOptionsFrame_OpenToCategory(self.ioptionsWidget.frame) end

function addon:OnEnable()
	local ACD = LibStub("AceConfigDialog-3.0")
	local GUI = LibStub("AceGUI-3.0")

	-- Reworked to use ideas from this thread:
	-- http://forums.wowace.com/showthread.php?t=17033
	--
	-- Rather than calling ACD:AddToBlizOptions(), we assemble a pair of frames
	-- (the options plus a custom one).  This means that the work that would have
	-- been done in AddToBlizOptions needs to be done by hand first.
	local blizframe = GUI:Create("BlizOptionsGroup")
	blizframe:SetName(nametag)
	blizframe:SetTitle(self.options.name)
	blizframe:SetLayout("List")
	blizframe:SetUserData('appName', nametag)
	ACD.BlizOptions[nametag] = blizframe

	blizframe:SetCallback("OnShow", function (bf, e, ...)
		local acdoptsframe = GUI:Create("SimpleGroup")    -- first frame, options table
		local currencyframe = GUI:Create("SimpleGroup")   -- second frame, currency grid
		acdoptsframe:SetFullWidth(true)
		currencyframe:SetFullWidth(true)
		acdoptsframe:SetHeight(250)              -- OPTIONS TABLE HEIGHT
		ACD:Open (nametag, acdoptsframe)         -- populates top frame with options table
		addon:FillOptionsWindow (currencyframe)  -- populates lower frame with our stuff
		blizframe:AddChild(acdoptsframe)
		blizframe:AddChild(currencyframe)
	end)

	blizframe:SetCallback("OnHide", function (bf, e, ...)
		bf:ReleaseChildren()
	end)

	self.ioptionsWidget = blizframe
	blizframe.frame.okay = options_set_informuser
	blizframe.frame.cancel = options_set_informuser -- this is technically an Ace bug/weakness, but a well-known one
	InterfaceOptions_AddCategory(blizframe.frame)
end

function addon:OnDisable()
end


-- GetMoney() returns 0 if called too quickly after PLAYER_LOGIN.  Sigh.
-- Waiting a few seconds should be enough.  The base UI happens to have a 15
-- second countdown starting at load time which we can hook into, which is
-- good enough for our purposes.
AutoCompleteInfoDelayer:HookScript("OnFinished", function()
	addon:Print("Updating money and alternate currency counts.")
	addon:gwReset()
	mySV.money = addon:gwBase()

	addon:RegisterEvent("PLAYER_MONEY")
--[==[
	--addon:RegisterEvent("MERCHANT_SHOW", "ScanCurrencies")
	--addon:RegisterEvent("MERCHANT_CLOSED", "ScanCurrencies")
	addon:RegisterEvent("MERCHANT_SHOW")
	addon:RegisterEvent("MERCHANT_CLOSED")
	--addon:RegisterEvent("PLAYER_LOGOUT", "ScanCurrencies")
	if TokenFrame then
		addon:ADDON_LOADED("Blizzard_TokenUI")  -- already loaded, catch up
	else
		addon:RegisterEvent("ADDON_LOADED")
	end
	--addon:ScanCurrencies()
	addon.not_yet_scanned = nil
]==]
end)


----- Tooltip handling
-- character/token counts
local tooltip
local QTip = LibStub("LibQTip-1.0")

local function get_tooltip()
	-- We use 3 columns instead of 2 because 
	-- http://www.wowace.com/addons/libqtip-1-0/tickets/8-right-edge-of-the-last-column-is-clipped/
	-- was closed without actually being fixed (apparently depends on fonts?)
	tooltip = QTip:Acquire(nametag, 3, "LEFT", "RIGHT")
	tooltip:SetFont(GameTooltipTextSmall)
	return tooltip
end
local function hide_tooltip()
	if tooltip then
		tooltip:SetParent(nil)
		QTip:Release(tooltip)
		tooltip = nil
	end
end

local function generic_currency_icon_onenter (itemID, opt_tooltip, opt_header,
                                              x_offset, y_offset)
	local tt = get_tooltip()
	local show_p = nil

	--tt:SmartAnchorTo(GameTooltip)
	local maintip = opt_tooltip or GameTooltip
	tt:SetParent(maintip)
	tt:SetPoint("TOPLEFT", maintip, "BOTTOMLEFT", x_offset or 40, y_offset or 0)

	if opt_header then
		tt:AddHeader(LIGHTYELLOW_FONT_COLOR_CODE..opt_header..FONT_COLOR_CODE_CLOSE)
		--tt:SetLineColor(1, PASSIVE_SPELL_FONT_COLOR.r, PASSIVE_SPELL_FONT_COLOR.g, PASSIVE_SPELL_FONT_COLOR.b)
	end

	for char,tokens in sorted_currency_pairs (addon.db.factionrealm, itemID) do
		show_p = tt:AddLine(char,tokens)
	end

	if show_p then tt:Show() end
	return tt
end
local function token_currency_setcurrency (gametooltip, currency_index)
	local id = token_map[currency_index]
	if id then generic_currency_icon_onenter(id) end
end
local function token_currency_setmerchantcostitem (gametooltip, index, item)
	local ilink = select(3,GetMerchantItemCostItem(index,item))
	local id = tonumber(ilink:match(".*item:(%d+):"))
	if id then generic_currency_icon_onenter(id) end
end
local function token_currency_icon_onenter (currency_icon)
	generic_currency_icon_onenter (currency_icon:GetParent().itemID)
end
local function options_currency_icon_onenter (widget)
	generic_currency_icon_onenter (widget:GetUserData("itemID"), widget.frame,
								   widget:GetUserData("token name"), 60)
end
--[[local function currency_icon_onleave()
	qtip:Release(tooltip)
	tooltip = nil
end]]


----- Console-related things
local function gwFormat(cp)
	return globalSV.useIcons and GetCoinTextureString(cp,globalSV.iconSize) or GetCoinText(cp, " ")
end

-- Handler for the /gw shortcut
function addon:gwSlash(txt)
	txt = strtrim(txt:lower())
	if txt == "all" or txt == "all+" then
		self:gwAll(txt == "all+")
	elseif txt == "reset" then
		self:gwReset()
		self:Print"GoldWatch delta counter reset"
	else
		self:gwStatus()
	end
end

-- Handler for the /tw shortcut
function addon:twSlash(txt)
	txt = strtrim(txt:lower())
	self:Print"This command does nothing yet."
end

local gwbasecp = 0
function addon:gwReset()
	gwbasecp = GetMoney()
end
function addon:gwBase()
	return gwbasecp
end

function addon:gwAll (extended_p, currency_type)
	if type(extended_p) == "table" then
		-- being called via /goldwatch, it's actually the 'info' table
		extended_p = extended_p.arg
	end
	if currency_type == nil then
		currency_type = 'money'
	end

	local count, total = 0, 0
	for toon,amount in sorted_currency_pairs(shareSV, currency_type) do
		count = count + 1
		total = total + amount
		if extended_p then
			self:Printf("%s: %s", toon, gwFormat(amount))  -- maybe insert some coloring codes later?
			--self:Print(k..": "..gwFormat(v))
		end
	end

	self:Printf("Total coin available to this character across %d toons is %s.",
			   count, gwFormat(total))
end

function addon:gwStatus ()
	local diff, verb = GetMoney() - self:gwBase(), "made"
	if diff < 0 then
		verb = "lost"
		diff = abs(diff)
	end
	self:Printf("You have %s %s.", verb, gwFormat(diff))
end


----- More GUI handling
-- Called as part of OnShow of the interface options window.  F is an AceGUI
-- container, probably scrollframe, with no parameters set.
function addon:FillOptionsWindow (f)
	local GUI = LibStub("AceGUI-3.0")

	--f:SetLayout("Fill")
	--f:SetFullWidth(true)

	if self.not_yet_scanned then
		local whoa = GUI:Create("Label")
		whoa:SetText("Currency token scan not yet finished.")
		whoa:SetImage("Interface\\DialogFrame\\DialogAlertIcon")
		whoa:SetImageSize(48,48)
		f:AddChild(whoa)
		return
	end

	local anytracked = next(globalSV.tracked)
	if not anytracked then
		f:SetFullWidth(false)
		f:SetWidth(150)
		local empty = GUI:Create("Label")
		empty:SetText(empty_text)
		empty:SetImage("Interface\\DialogFrame\\DialogAlertIcon")
		f:AddChild(empty)
		local currency = GUI:Create("Button")
		currency:SetText("Open Currency Window")
		currency:SetCallback("OnClick", function()
			InterfaceOptionsFrame:Hide()
			ToggleCharacter("TokenFrame")
		end)
		f:AddChild(currency)
		return
	end

	local scroll = GUI:Create("ScrollFrame")
	scroll:SetLayout("List")
	scroll:SetFullWidth(true)
	scroll:SetHeight(100)

	for id in pairs(globalSV.tracked) do
		local name = GetItemInfo(id)
		local icon = GUI:Create("InteractiveLabel")
		icon:SetWidth(300)
		icon:SetHeight(28)
		icon:SetImage(smartGetItemIcon(id))
		icon:SetImageSize(18,18)
		--icon:SetLabel(name)
		icon:SetText(name)
		icon:SetUserData("token name",name)
		icon:SetUserData("itemID",id)
		icon:SetCallback("OnEnter", options_currency_icon_onenter)
		icon:SetCallback("OnLeave", hide_tooltip)
		scroll:AddChild(icon)
		if false then  -- double up the list for testing the scrollbar
			local testicon = GUI:Create("InteractiveLabel")
			testicon:SetWidth(700)
			testicon:SetHeight(25)
			testicon:SetImage(smartGetItemIcon(id))
			testicon:SetImageSize(15,15)
			--testicon:SetLabel(name)
			testicon:SetText(name)
			testicon:SetUserData("itemID",id)
			testicon:SetCallback("OnEnter", options_currency_icon_onenter)
			testicon:SetCallback("OnLeave", hide_tooltip)
			scroll:AddChild(testicon)
		end
	end

	f:AddChild(scroll)
end



















--[[
Section headers are broken out like this, for easy searching:
----- Tooltip handling
----- Massive options collection
----- Console-related things
----- Currency token tracking
----- Ace framework hooks
----- More GUI handling
----- Event handlers
]]

















----- Currency token tracking
-- Updates current character's "alternate currency" counts.  This is probably
-- going to be called very often, since there are many cases where the "real"
-- number might change but does not generate an event.  If called as an event
-- handler, EVENT will be the name; if called directly it will be nil.
--[==[
function addon:ScanCurrencies(event,...)
	--print("Scan running",event,...)
	for i = 1, GetCurrencyListSize() do
		-- name, isHeader, isExpanded, isUnused, isWatched, count, extraCurrencyType, icon, itemID
		local count,ct,_,id = select(6,GetCurrencyListInfo(i))
		if globalSV.tracked[id] then
			mySV[id] = count
			token_map[i] = id
			cache_smartGetItemIcon(id,ct)
		end
	end

	if (event == nil) then --and self.ioptionsFrame:IsVisible() then
		--LibStub("AceConfigRegis
	end
end

function addon:addTokenTrack (link, id)
	globalSV.tracked[id] = true
	self:Print(("Added %s to currency tracking.  Counts will be updated as characters are logged on."):format(link))
end

function addon:removeTokenTrack (link, id)
	globalSV.tracked[id] = nil
	self:Print(("Removed %s from currency tracking."):format(link))
end
]==]






----- Event handlers
--[==[
function addon:ADDON_LOADED(which)
	if which ~= "Blizzard_TokenUI" then return end

	-- Additional checkbox in the currency tab popout.
	local cb = CreateFrame("CheckButton", "GoldWatchTokenFrameCheckBox",
						   TokenFramePopup, "OptionsSmallCheckButtonTemplate")
	GoldWatchTokenFrameCheckBoxText:SetText("Track in Gold Watch")
	GoldWatchTokenFrameCheckBoxText:SetTextColor(HIGHLIGHT_FONT_COLOR.r,
									 HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	cb:SetPoint("TOPLEFT", TokenFramePopupBackpackCheckBox, "BOTTOMLEFT", 0, 4)
	TokenFramePopup:SetHeight(TokenFramePopup:GetHeight() + 13)
	cb:SetScript("OnEnter", function(box)
		GameTooltip:SetOwner(box, "ANCHOR_RIGHT")
		GameTooltip:SetText(currency_checkbox_tooltip_text, nil, nil, nil, nil, --[[wrap=]]1)
	end)
	cb:SetScript("OnLeave", GameTooltip_Hide)
	cb:SetScript("OnHide", function(box)
		box.currencybutton = nil
	end)
	cb:SetScript("OnClick", function(box)
		local id = box.currencybutton.itemID
		local linktext = select(2, GetItemInfo(id))
		if box:GetChecked() then
			PlaySound("igMainMenuOptionCheckBoxOn")
			addon:addTokenTrack(linktext, id)
			addon:ScanCurrencies()
		else
			PlaySound("igMainMenuOptionCheckBoxOff")
			addon:removeTokenTrack(linktext, id)
		end
	end)

	-- orig in Blizzard_TokenUI.lua, updates the normal two checkboxes
	hooksecurefunc("TokenFrame_UpdatePopup", function (currency)
		cb:SetChecked(globalSV.tracked[currency.itemID])
		cb.currencybutton = currency
	end)

	hooksecurefunc(GameTooltip, "SetCurrencyToken", token_currency_setcurrency)
	hooksecurefunc(GameTooltip, "SetMerchantCostItem", token_currency_setmerchantcostitem)
	TokenFrame:HookScript("OnShow", self.MERCHANT_SHOW)
	TokenFrame:HookScript("OnHide", self.MERCHANT_CLOSED)
--[[
	-- Hook the mouseover of currency "buttons".  What we want to hook, ideally,
	-- is OnEnter from TokenButtonTemplate or TokenButtonTemplate.LinkButton, but
	-- virtuals don't exist.  Instead, we grab the first "button" (row) of the
	-- currency window and hook that one.  Running the hook over all the buttons
	-- replaces a single function with N identical copies of the hooked function
	-- (sigh), so after setting up the first hook, we directly set the remaining
	-- N-1 handlers to the new hooked function.  Ditto all that for OnLeave.
	local buttons = TokenFrameContainer.buttons
	local b1 = buttons and buttons[1]
	if b1 then
		b1.LinkButton:HookScript("OnEnter", token_currency_icon_onenter)
		b1.LinkButton:HookScript("OnLeave", hide_tooltip) --currency_icon_onleave)
		local newenter, newleave = b1.LinkButton:GetScript("OnEnter"), b1.LinkButton:GetScript("OnLeave")
		for i,b in ipairs(buttons) do
			b.LinkButton:SetScript("OnEnter", newenter)
			b.LinkButton:SetScript("OnLeave", newleave)
		end
	--else ???  self:OnDisable()?
	end
]]

	self:UnregisterEvent("ADDON_LOADED")
	self.ADDON_LOADED = nil
end
]==]

function addon:PLAYER_MONEY()
	mySV.money = GetMoney()
end

--[==[
local gt_onhide = GameTooltip_OnHide
local gt_hooked = 0
local function hooked_gt_onhide(self)
	hide_tooltip()
	--gt_onhide(self)
end

-- N.B.:  These are also [ab]used as scripts for other frames, so do not
-- use 'self' here.
function addon:MERCHANT_SHOW()
	if gt_hooked == 0 then
		--GameTooltip_OnHide = hooked_gt_onhide
		addon:SecureHook ("GameTooltip_OnHide", hooked_gt_onhide)
	end
	gt_hooked = gt_hooked + 1
	addon:ScanCurrencies()
end

function addon:MERCHANT_CLOSED()
	gt_hooked = gt_hooked - 1
	if gt_hooked < 1 then
		gt_hooked = 0
		--GameTooltip_OnHide = gt_onhide
		addon:Unhook ("GameTooltip_OnHide")
	end
	addon:ScanCurrencies()
end
]==]

-- vim: noet
